// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/24DSI16WRC/driver/ioctl.c $
// $Rev: 52515 $
// $Date: 2023-03-01 12:50:28 -0600 (Wed, 01 Mar 2023) $

// 24DSI16WRC: Device Driver: source file

#include "main.h"



// macros *********************************************************************

#define	AUTO_CAL_START			D7		// BCTLR
#define	AUTO_CAL_PASS			D12		// BCTLR
#define	INIT_START				D15		// BCTLR

#define	IRQ_INIT_DONE			0x0000	// BCTLR
#define	IRQ_AUTO_CAL_DONE		0x0100	// BCTLR
#define	IRQ_MASK				0x0700	// BCTLR
#define	IRQ_ACTIVE				0x0800	// BCTLR



// data types *****************************************************************

typedef struct
{
	int	sw;
	int	hw;
} sw_hw_t;



// prototypes *****************************************************************

static	int		_wait_for_ready(dev_data_t* dev, const char* reason, long ms_limit);
static	int		_xlat_hw_2_sw(s32* arg, const sw_hw_t* list);
static	int		_xlat_sw_2_hw(s32* arg, const sw_hw_t* list);



//*****************************************************************************
static void _bctlr_mod(dev_data_t* dev, u32 value, u32 mask)
{
	if (mask != IRQ_ACTIVE)
	{
		value	|= IRQ_ACTIVE;
		mask	|= IRQ_ACTIVE;
	}

	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bctlr_32, value, mask);
}



//*****************************************************************************
static int _query(dev_data_t* dev, void* arg)
{
	int		ret	= 0;
	s32*	ptr	= (s32*) arg;

	switch (ptr[0])
	{
		default:	ptr[0]	= DSI16WRC_IOCTL_QUERY_ERROR;
					ret		= -EINVAL;
					break;

		case DSI16WRC_QUERY_AUTO_CAL_MS:	ptr[0]	= dev->cache.auto_cal_ms;	break;
		case DSI16WRC_QUERY_CHANNEL_MAX:	ptr[0]	= dev->cache.channels_max;	break;
		case DSI16WRC_QUERY_CHANNEL_QTY:	ptr[0]	= dev->cache.channel_qty;	break;
		case DSI16WRC_QUERY_COUNT:			ptr[0]	= DSI16WRC_QUERY_LAST - 1;	break;
		case DSI16WRC_QUERY_DEVICE_TYPE:	ptr[0]	= dev->board_type;			break;
		case DSI16WRC_QUERY_EXT_SYNC_IN:	ptr[0]	= dev->cache.ext_sync_in;	break;
		case DSI16WRC_QUERY_FGEN_MAX:		ptr[0]	= dev->cache.fgen_max;		break;
		case DSI16WRC_QUERY_FGEN_MIN:		ptr[0]	= dev->cache.fgen_min;		break;
		case DSI16WRC_QUERY_FIFO_SIZE:		ptr[0]	= dev->cache.fifo_size;		break;
		case DSI16WRC_QUERY_FILTER_FREQ:	ptr[0]	= dev->cache.filter_freq;	break;
		case DSI16WRC_QUERY_FREF_DEFAULT:	ptr[0]	= dev->cache.fref_default;	break;
		case DSI16WRC_QUERY_FSAMP_MAX:		ptr[0]	= dev->cache.fsamp_max;		break;
		case DSI16WRC_QUERY_FSAMP_MIN:		ptr[0]	= dev->cache.fsamp_min;		break;
		case DSI16WRC_QUERY_INIT_MS:		ptr[0]	= dev->cache.initialize_ms;	break;
		case DSI16WRC_QUERY_NDIV_MAX:		ptr[0]	= dev->cache.ndiv_max;		break;
		case DSI16WRC_QUERY_NDIV_MIN:		ptr[0]	= dev->cache.ndiv_min;		break;
		case DSI16WRC_QUERY_NREF_MAX:		ptr[0]	= dev->cache.nref_max;		break;
		case DSI16WRC_QUERY_NREF_MIN:		ptr[0]	= dev->cache.nref_min;		break;
		case DSI16WRC_QUERY_NVCO_MAX:		ptr[0]	= dev->cache.nvco_max;		break;
		case DSI16WRC_QUERY_NVCO_MIN:		ptr[0]	= dev->cache.nvco_min;		break;
		case DSI16WRC_QUERY_V_RANGE:		ptr[0]	= dev->cache.voltage_range;	break;
	}

	return(ret);
}



//*****************************************************************************
static int _init_start(dev_data_t* dev, void* arg)
{
	// Initiate initialization.
	_bctlr_mod(dev, INIT_START, INIT_START);
	return(0);
}



//*****************************************************************************
int initialize_ioctl(dev_data_t* dev, void* arg)
{
	u32			bctlr;
	int			i;
	int			mask;
	long		ms			= dev->cache.initialize_ms + 5000;
	long		ms_total	= ms;
	os_sem_t	sem;
	int			ret			= 0;
	int			tmp;
	int			value;
	gsc_wait_t	wt;

	// In the original firmware the INIT bit clears right away, while the
	// interrupt is generated when initialization is completed. As of firmware
	// version 00A the bit clears when initialization is done. We'll use the
	// initerrupt to detect completion as it works with all versions.

	if (dev->irq.opened)
	{
		ms_total	*= 2;

		// Safely select the Initialize Done interrupt.
		mask	= IRQ_MASK | IRQ_ACTIVE;
		value	= IRQ_INIT_DONE;
		os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bctlr_32, value, mask);

		// Wait for the local interrupt.
		os_sem_create(&sem);	// dummy, required for wait operations.
		memset(&wt, 0, sizeof(wt));
		wt.flags		= GSC_WAIT_FLAG_INTERNAL;
		wt.gsc			= DSI16WRC_WAIT_GSC_INIT_DONE;
		wt.timeout_ms	= ms;
		ret				= gsc_wait_event(dev, &wt, _init_start, NULL, &sem);
		os_sem_destroy(&sem);

		if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
		{
			ret	= ret ? ret : -ETIMEDOUT;
			printf(	"%s: INITILIZE DONE IRQ TIMED OUT AFTER %ld ms.\n",
					dev->model,
					(long) ms);
		}
	}

	// Manually wait for completion as the IRQ wait may have ended early.
	tmp	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bctlr_32, INIT_START, 0);

	if (tmp)
	{
		ret	= ret ? ret : tmp;
		printf(	"%s: INITIALIZATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms_total);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	bctlr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bctlr_32);

	if (bctlr & INIT_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: INITILIZE STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms_total);
	}

	// Initialize the software settings.

	for (i = 0; i < DEV_IO_STREAM_QTY; i++)
	{
		if (dev->io.io_streams[i])
		{
			if (dev->io.io_streams[i]->dev_io_sw_init)
			{
				(dev->io.io_streams[i]->dev_io_sw_init)(dev, dev->io.io_streams[i]);
			}
		}
	}

	return(ret);
}



//*****************************************************************************
static int _auto_cal_start(dev_data_t* dev, void* arg)
{
	// Initiate auto-calibration.
	_bctlr_mod(dev, AUTO_CAL_START, AUTO_CAL_START);
	return(0);
}



//*****************************************************************************
static int _auto_calibrate(dev_data_t* dev, void* arg)
{
	u32			bctlr;
	u32			mask;
	long		ms		= dev->cache.auto_cal_ms + 5000;
	u32			reg;
	int			ret;
	os_sem_t	sem;
	int			tmp;
	u32			value;
	gsc_wait_t	wt;

	// Safely select the Auto-Cal Done interrupt.
	mask	= IRQ_MASK | IRQ_ACTIVE;
	value	= IRQ_AUTO_CAL_DONE;
	_bctlr_mod(dev, value, mask);

	// Wait for the local interrupt.
	os_sem_create(&sem);	// dummy, required for wait operations.
	memset(&wt, 0, sizeof(wt));
	wt.flags		= GSC_WAIT_FLAG_INTERNAL;
	wt.gsc			= DSI16WRC_WAIT_GSC_AUTO_CAL_DONE;
	wt.timeout_ms	= ms;
	ret				= gsc_wait_event(dev, &wt, _auto_cal_start, NULL, &sem);
	os_sem_destroy(&sem);

	if (wt.flags & GSC_WAIT_FLAG_TIMEOUT)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATE IRQ TIMED OUT AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually wait for completion in case something terminates our wait early.
	tmp	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bctlr_32, AUTO_CAL_START, 0);

	if (tmp)
	{
		ms	*= 2;
		ret	= ret ? ret : tmp;
		printf(	"%s: AUTOCALIBRATION DID NOT COMPLETE WITHIN %ld ms.\n",
				dev->model,
				ms);
	}

	// Manually check for completion as the IRQ wait may have ended early.
	bctlr	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bctlr_32);

	if (bctlr & AUTO_CAL_START)
	{
		ret	= ret ? ret : -ETIMEDOUT;
		printf(	"%s: AUTOCALIBRATION STILL ACTIVE AFTER %ld ms.\n",
				dev->model,
				ms);
	}

	// Final results.
	reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bctlr_32);

	if ((reg & AUTO_CAL_PASS) == 0)
	{
		ret	= ret ? ret : -EIO;
		printf(	"%s: AUTOCALIBRATION FAILED (%ld ms).\n",
				dev->model,
				ms);
	}

	if (ret == 0)
	{
		// Wait for settling.
		os_time_sleep_ms(100);
	}

	return(ret);
}



//*****************************************************************************
static int _auto_cal_sts(dev_data_t* dev, s32* arg)
{
	u32	reg;

	reg	= os_reg_mem_rx_u32(dev, dev->vaddr.gsc_bctlr_32);

	if (reg & AUTO_CAL_START)
		arg[0]	= DSI16WRC_AUTO_CAL_STS_ACTIVE;
	else if (reg & AUTO_CAL_PASS)
		arg[0]	= DSI16WRC_AUTO_CAL_STS_PASS;
	else
		arg[0]	= DSI16WRC_AUTO_CAL_STS_FAIL;

	return(0);
}



//*****************************************************************************
static int _adc_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_ADC_MODE_HI_RES,
		DSI16WRC_ADC_MODE_HI_SPEED,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 19, 19);

	// Wait for settling.

	if (ret  == 0)
		ret	= _wait_for_ready(dev, "Hi Speed/Hi Res Change", 800);

	return(ret);
}



//*****************************************************************************
static int _ai_buf_clear(dev_data_t* dev, void* arg)
{
	int	ms		= 50;	// Manual says 20ms max.
	int	ret;

	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bufcr_32, D20, D20 | D24 | D25);

	// Wait for the bit to clear.
	ret	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bufcr_32, D20, 0);

	if (ret)
	{
		ret	= -ETIMEDOUT;
		printf(	"%s: AI BUFFER CLEAR DID NOT COMPLETE WITHIN %d ms.\n",
				dev->model,
				ms);
	}

	// Wait for settling.

	if (ret == 0)
		ret	= _wait_for_ready(dev, "AI Buffer Clear", 25);

	return(ret);
}



//*****************************************************************************
static int _ai_buf_enable(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_AI_BUF_ENABLE_NO,
		DSI16WRC_AI_BUF_ENABLE_YES,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bufcr_32, 19, 19);
	return(ret);
}



//*****************************************************************************
static int _ai_buf_fill_lvl(dev_data_t* dev, s32* arg)
{
	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_range_reg(dev, arg, 0, 0x40000, dev->vaddr.gsc_bufsr_32, 18, 0);
	return(ret);
}



//*****************************************************************************
static int _ai_buf_overflow(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_AI_BUF_UNDERFLOW_NO,
		DSI16WRC_AI_BUF_UNDERFLOW_YES,
		-1	// terminate list
	};

	int	ret;

	if (arg[0] == 1)
		ret	= -EINVAL;
	else
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bufcr_32, 24, 24);

	return(ret);
}



//*****************************************************************************
static int _ai_buf_thresh(dev_data_t* dev, void* arg)
{
	int	ret;

	ret		= gsc_s32_range_reg(dev, arg, 0, 0x40000, dev->vaddr.gsc_bufcr_32, 18, 0);
	return(ret);
}



//*****************************************************************************
static int _ai_buf_thr_sts(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_AI_BUF_THR_STS_IDLE,
		DSI16WRC_AI_BUF_THR_STS_ACTIVE,
		-1	// terminate list
	};

	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 14, 14);
	return(ret);
}



//*****************************************************************************
static int _ai_buf_underflow(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_AI_BUF_OVERFLOW_NO,
		DSI16WRC_AI_BUF_OVERFLOW_YES,
		-1	// terminate list
	};

	int	ret;

	if (arg[0] == 1)
		ret	= -EINVAL;
	else
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bufcr_32, 25, 25);

	return(ret);
}



//*****************************************************************************
static int _ai_channel_tag(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_AI_CHANNEL_TAG_ENABLE,
		DSI16WRC_AI_CHANNEL_TAG_DISABLE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bufcr_32, 23, 23);
	return(ret);
}



//*****************************************************************************
static int _ai_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_AI_MODE_DIFF,
		DSI16WRC_AI_MODE_ZERO,
		DSI16WRC_AI_MODE_VREF,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 1, 0);
	return(ret);
}



//*****************************************************************************
static int _aux_clk_ctl_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_AUX_CLK_CTL_MODE_INACTIVE,
		DSI16WRC_AUX_CLK_CTL_MODE_INPUT,
		DSI16WRC_AUX_CLK_CTL_MODE_OUTPUT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 1, 0);
	return(ret);
}



//*****************************************************************************
static int _aux_sync_ctl_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_AUX_SYNC_CTL_MODE_INACTIVE,
		DSI16WRC_AUX_SYNC_CTL_MODE_INPUT,
		DSI16WRC_AUX_SYNC_CTL_MODE_OUTPUT,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_asiocr_32, 3, 2);
	return(ret);
}



//*****************************************************************************
static int _burst(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_BURST_DISABLE,
		DSI16WRC_BURST_ENABLE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 21, 21);
	return(ret);
}



//*****************************************************************************
static int _burst_rate_div(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xFFFFFF, dev->vaddr.gsc_bttr_32, 23, 0);
	return(ret);
}



//*****************************************************************************
static int _burst_size(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xFFFFFF, dev->vaddr.gsc_bbsr_32, 23, 0);
	return(ret);
}



//*****************************************************************************
static int _burst_timer(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_BURST_TIMER_DISABLE,
		DSI16WRC_BURST_TIMER_ENABLE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 16, 16);
	return(ret);
}



//*****************************************************************************
static int _burst_trigger(dev_data_t* dev, void* arg)
{
	os_reg_mem_mx_u32(dev, dev->vaddr.gsc_bctlr_32, D22, D22);
	return(0);
}



//*****************************************************************************
static int _ch_grp_0_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_CH_GRP_SRC_RATE_GEN,
		DSI16WRC_CH_GRP_SRC_EXTERN,
		DSI16WRC_CH_GRP_SRC_DIR_EXTERN,
		DSI16WRC_CH_GRP_SRC_DISABLE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_csar_32, 3, 0);

	// Wait for settling.

	if (ret == 0)
		ret	= _wait_for_ready(dev, "Channel Group 0 source change", 25);

	return(ret);
}



//*****************************************************************************
static int _ch_grp_1_src(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_CH_GRP_SRC_RATE_GEN,
		DSI16WRC_CH_GRP_SRC_EXTERN,
		DSI16WRC_CH_GRP_SRC_DIR_EXTERN,
		DSI16WRC_CH_GRP_SRC_DISABLE,
		-1	// terminate list
	};

	s32	g0	= -1;
	s32	g1	= -1;
	int	ret;
	int	ret0;
	int	ret1;

	if (arg[0] == -1)
	{
		ret0	= gsc_s32_list_reg(dev, &g0, options, dev->vaddr.gsc_csar_32, 3, 0);
		ret1	= gsc_s32_list_reg(dev, &g1, options, dev->vaddr.gsc_csar_32, 7, 4);

		arg[0]	= (g1 == DSI16WRC_CH_GRP_SRC_DISABLE) ? g1 : g0;
		ret		= ret0 ? ret0 : ret1;
	}
	else
	{
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_csar_32, 7, 4);
	}

	// Wait for settling.

	if (ret == 0)
		ret	= _wait_for_ready(dev, "Channel Group 0 and 1 source change", 25);

	return(ret);
}



//*****************************************************************************
static int _channels_ready(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_CHANNELS_READY_NO,
		DSI16WRC_CHANNELS_READY_YES,
		-1	// terminate list
	};

	int	ret	= 0;

	if (arg[0] == DSI16WRC_CHANNELS_READY_WAIT)
	{
		gsc_poll_u32(dev, 1000, dev->vaddr.gsc_bctlr_32, D13, D13);
		arg[0]	= DSI16WRC_CHANNELS_READY_TEST;
	}

	if (arg[0] == DSI16WRC_CHANNELS_READY_TEST)
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 13, 13);
	else
		ret	= -EINVAL;

	return(ret);
}



//*****************************************************************************
static int _control_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_CONTROL_MODE_TARGET,
		DSI16WRC_CONTROL_MODE_INITIATOR,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 5, 5);
	return(ret);
}



//*****************************************************************************
static int _data_format(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_DATA_FORMAT_2S_COMP,
		DSI16WRC_DATA_FORMAT_OFF_BIN,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 4, 4);
	return(ret);
}



//*****************************************************************************
static int _data_width(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_DATA_WIDTH_16,
		DSI16WRC_DATA_WIDTH_18,
		DSI16WRC_DATA_WIDTH_20,
		DSI16WRC_DATA_WIDTH_24,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bufcr_32, 22, 21);
	return(ret);
}



//*****************************************************************************
static int _dio_dir_out(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xF, dev->vaddr.gsc_diopr_32, 11, 8);
	return(ret);
}



//*****************************************************************************
static int _dio_read(dev_data_t* dev, s32* arg)
{
	int	ret;

	arg[0]	= -1;
	ret		= gsc_s32_range_reg(dev, arg, 0, 0xF, dev->vaddr.gsc_diopr_32, 3, 0);
	return(ret);
}



//*****************************************************************************
static int _dio_write(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 0, 0xF, dev->vaddr.gsc_diopr_32, 3, 0);
	return(ret);
}



//*****************************************************************************
static int _ext_clk_src(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_EXT_CLK_SRC_GRP_0,
		DSI16WRC_EXT_CLK_SRC_GEN,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 18, 18);
	return(ret);
}



//*****************************************************************************
static int _irq_sel(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_IRQ_INIT_DONE,
		DSI16WRC_IRQ_AUTO_CAL_DONE,
		DSI16WRC_IRQ_CHAN_READY,
		DSI16WRC_IRQ_AI_BUF_THRESH_L2H,
		DSI16WRC_IRQ_AI_BUF_THRESH_H2L,
		DSI16WRC_IRQ_AI_BURST_DONE,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 10, 8);
	return(ret);
}



//*****************************************************************************
static int _master_clk_adj(dev_data_t* dev, void* arg)
{
	int	ret;

	ret		= gsc_s32_range_reg(dev, arg, 0, 0xFFFF, dev->vaddr.gsc_mcar_32, 15, 0);
	return(ret);
}



//*****************************************************************************
static int _ndiv(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 1, 300, dev->vaddr.gsc_rdr_32, 8, 0);

	// Wait for settling.

	if (ret == 0)
		ret	= _wait_for_ready(dev, "Ndiv change", 25);

	return(ret);
}



//*****************************************************************************
static int _nref(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 25, 300, dev->vaddr.gsc_rcr_32, 23, 12);

	// Wait for settling.

	if (ret == 0)
		ret	= _wait_for_ready(dev, "Nref change", 25);

	return(ret);
}



//*****************************************************************************
static int _nvco(dev_data_t* dev, void* arg)
{
	int	ret;

	ret	= gsc_s32_range_reg(dev, arg, 25, 300, dev->vaddr.gsc_rcr_32, 11, 0);


	// Wait for settling.

	if (ret == 0)
		ret	= _wait_for_ready(dev, "Nvco change", 25);

	return(ret);
}



//*****************************************************************************
static int _range(dev_data_t* dev, s32* arg)
{
	static const sw_hw_t	list_10_by_2[]	=
	{
		// sw					hw
		{ DSI16WRC_RANGE_1_25V,	0	},
		{ DSI16WRC_RANGE_2_5V,	1	},
		{ DSI16WRC_RANGE_5V,	2	},
		{ DSI16WRC_RANGE_10V,	3	},
		{ -1, -1, }
	};

	static const sw_hw_t	list_10_by_10[]	=
	{
		// sw					hw
		{ DSI16WRC_RANGE_10MV,	0	},
		{ DSI16WRC_RANGE_100MV,	1	},
		{ DSI16WRC_RANGE_1V,	2	},
		{ DSI16WRC_RANGE_10V,	3	},
		{ -1, -1, }
	};

	const sw_hw_t*	list;
	int				ret	= 0;

	if (dev->cache.voltage_range == DSI16WRC_QUERY_V_RANGE_10_BY_2)
		list	= list_10_by_2;
	else
		list	= list_10_by_10;

	if (arg[0] == -1)
	{
		ret	= gsc_s32_range_reg(dev, arg, 0, 3, dev->vaddr.gsc_bctlr_32, 3, 2);
		_xlat_hw_2_sw(arg, list);
	}
	else
	{
		ret	= _xlat_sw_2_hw(arg, list);

		if (ret == 0)
			gsc_s32_range_reg(dev, arg, 0, 3, dev->vaddr.gsc_bctlr_32, 3, 2);

		_xlat_hw_2_sw(arg, list);
	}

	return(ret);
}



//*****************************************************************************
static int _rx_io_abort(dev_data_t* dev, s32* arg)
{
	arg[0]	= gsc_read_abort_active_xfer(dev, &dev->io.rx);
	return(0);
}



//*****************************************************************************
static int _rx_io_mode(dev_data_t* dev, s32* arg)
{
	static const s32	list[]	=
	{
		GSC_IO_MODE_PIO,
		GSC_IO_MODE_BMDMA,
		GSC_IO_MODE_DMDMA,
		-1
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, list, &dev->io.rx.io_mode);
	return(ret);
}



//*****************************************************************************
static int _rx_io_overflow(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_IO_OVERFLOW_IGNORE,
		DSI16WRC_IO_OVERFLOW_CHECK,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.rx.overflow_check);
	return(ret);
}



//*****************************************************************************
static int _rx_io_timeout(dev_data_t* dev, s32* arg)
{
	int	ret;

	ret	= gsc_s32_range_var(
			arg,
			DSI16WRC_IO_TIMEOUT_MIN,
			DSI16WRC_IO_TIMEOUT_INFINITE,
			&dev->io.rx.timeout_s);
	return(ret);
}



//*****************************************************************************
static int _rx_io_underflow(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_IO_UNDERFLOW_IGNORE,
		DSI16WRC_IO_UNDERFLOW_CHECK,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_var(arg, options, &dev->io.rx.underflow_check);
	return(ret);
}



//*****************************************************************************
static int _sw_sync(dev_data_t* dev, void* arg)
{
	int	ms		= 1100;
	int	ret;

	// The device manual says a sync operation should take 700 ms or less, but
	// that it could take as long as a second for very low sample rates.

	_bctlr_mod(dev, D6, D6);

	// Wait for the bit to clear.
	ret	= gsc_poll_u32(dev, ms, dev->vaddr.gsc_bctlr_32, D6, 0);

	if (ret)
	{
		ret	= -ETIMEDOUT;
		printf(	"%s: SOFTWARE SYNC DID NOT COMPLETE WITHIN %d ms.\n",
				dev->model,
				ms);
	}

	if (ret  == 0)
		ret	= _wait_for_ready(dev, "Software SYNC", ms);

	return(ret);
}



//*****************************************************************************
static int _sw_sync_mode(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_SW_SYNC_MODE_SYNC,
		DSI16WRC_SW_SYNC_MODE_CLR_BUF,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 17, 17);
	return(ret);
}



//*****************************************************************************
static int _xcvr_type(dev_data_t* dev, void* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_XCVR_TYPE_LVDS,
		DSI16WRC_XCVR_TYPE_TTL,
		-1	// terminate list
	};

	int	ret;

	ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 20, 20);
	return(ret);
}


//*****************************************************************************
static int _ext_sync_in(dev_data_t* dev, s32* arg)
{
	static const s32	options[]	=
	{
		DSI16WRC_EXT_SYNC_IN_DISABLE,
		DSI16WRC_EXT_SYNC_IN_ENABLE,
		-1	// terminate list
	};

	int	ret;

	if (dev->cache.ext_sync_in)
	{
		ret	= gsc_s32_list_reg(dev, arg, options, dev->vaddr.gsc_bctlr_32, 23, 23);
	}
	else
	{
		switch (arg[0])
		{
			default:

				ret	= -EINVAL;
				break;

			case -1:
			case DSI16WRC_EXT_SYNC_IN_DISABLE:
			case DSI16WRC_EXT_SYNC_IN_ENABLE:

				arg[0]	= -1;
				ret		= 0;
				break;
		}
	}

	return(ret);
}



// variables ******************************************************************

const gsc_ioctl_t	dev_ioctl_list[]	=
{
	{ DSI16WRC_IOCTL_REG_READ,			(void*) gsc_reg_read_ioctl		},
	{ DSI16WRC_IOCTL_REG_WRITE,			(void*) gsc_reg_write_ioctl		},
	{ DSI16WRC_IOCTL_REG_MOD,			(void*) gsc_reg_mod_ioctl		},
	{ DSI16WRC_IOCTL_QUERY,				(void*) _query					},
	{ DSI16WRC_IOCTL_INITIALIZE,		(void*) initialize_ioctl		},
	{ DSI16WRC_IOCTL_AUTO_CALIBRATE,	(void*) _auto_calibrate			},
	{ DSI16WRC_IOCTL_AUTO_CAL_STS,		(void*) _auto_cal_sts			},
	{ DSI16WRC_IOCTL_ADC_MODE,			(void*) _adc_mode				},
	{ DSI16WRC_IOCTL_AI_BUF_CLEAR,		(void*) _ai_buf_clear			},
	{ DSI16WRC_IOCTL_AI_BUF_ENABLE,		(void*) _ai_buf_enable			},
	{ DSI16WRC_IOCTL_AI_BUF_FILL_LVL,	(void*) _ai_buf_fill_lvl		},
	{ DSI16WRC_IOCTL_AI_BUF_OVERFLOW,	(void*) _ai_buf_overflow		},
	{ DSI16WRC_IOCTL_AI_BUF_THRESH,		(void*) _ai_buf_thresh			},
	{ DSI16WRC_IOCTL_AI_BUF_THR_STS,	(void*) _ai_buf_thr_sts			},
	{ DSI16WRC_IOCTL_AI_BUF_UNDERFLOW,	(void*) _ai_buf_underflow		},
	{ DSI16WRC_IOCTL_AI_CHANNEL_TAG,	(void*) _ai_channel_tag			},
	{ DSI16WRC_IOCTL_AI_MODE,			(void*) _ai_mode				},
	{ DSI16WRC_IOCTL_AUX_CLK_CTL_MODE,	(void*) _aux_clk_ctl_mode		},
	{ DSI16WRC_IOCTL_AUX_SYNC_CTL_MODE,	(void*) _aux_sync_ctl_mode		},
	{ DSI16WRC_IOCTL_BURST,				(void*) _burst					},
	{ DSI16WRC_IOCTL_BURST_RATE_DIV,	(void*) _burst_rate_div			},
	{ DSI16WRC_IOCTL_BURST_SIZE,		(void*) _burst_size				},
	{ DSI16WRC_IOCTL_BURST_TIMER,		(void*) _burst_timer			},
	{ DSI16WRC_IOCTL_BURST_TRIGGER,		(void*) _burst_trigger			},
	{ DSI16WRC_IOCTL_CH_GRP_0_SRC,		(void*) _ch_grp_0_src			},
	{ DSI16WRC_IOCTL_CH_GRP_1_SRC,		(void*) _ch_grp_1_src			},
	{ DSI16WRC_IOCTL_CHANNELS_READY,	(void*) _channels_ready			},
	{ DSI16WRC_IOCTL_CONTROL_MODE,		(void*) _control_mode			},
	{ DSI16WRC_IOCTL_DATA_FORMAT,		(void*) _data_format			},
	{ DSI16WRC_IOCTL_DATA_WIDTH,		(void*) _data_width				},
	{ DSI16WRC_IOCTL_DIO_DIR_OUT,		(void*) _dio_dir_out			},
	{ DSI16WRC_IOCTL_DIO_READ,			(void*) _dio_read				},
	{ DSI16WRC_IOCTL_DIO_WRITE,			(void*) _dio_write				},
	{ DSI16WRC_IOCTL_EXT_CLK_SRC,		(void*) _ext_clk_src			},
	{ DSI16WRC_IOCTL_IRQ_SEL,			(void*) _irq_sel				},
	{ DSI16WRC_IOCTL_MASTER_CLK_ADJ,	(void*) _master_clk_adj			},
	{ DSI16WRC_IOCTL_NDIV,				(void*) _ndiv					},
	{ DSI16WRC_IOCTL_NREF,				(void*) _nref					},
	{ DSI16WRC_IOCTL_NVCO,				(void*) _nvco					},
	{ DSI16WRC_IOCTL_RANGE,				(void*) _range					},
	{ DSI16WRC_IOCTL_RX_IO_ABORT,		(void*) _rx_io_abort			},
	{ DSI16WRC_IOCTL_RX_IO_MODE,		(void*) _rx_io_mode				},
	{ DSI16WRC_IOCTL_RX_IO_OVERFLOW,	(void*) _rx_io_overflow			},
	{ DSI16WRC_IOCTL_RX_IO_TIMEOUT,		(void*) _rx_io_timeout			},
	{ DSI16WRC_IOCTL_RX_IO_UNDERFLOW,	(void*) _rx_io_underflow		},
	{ DSI16WRC_IOCTL_SW_SYNC,			(void*) _sw_sync				},
	{ DSI16WRC_IOCTL_SW_SYNC_MODE,		(void*) _sw_sync_mode			},
	{ DSI16WRC_IOCTL_WAIT_EVENT,		(void*) gsc_wait_event_ioctl	},
	{ DSI16WRC_IOCTL_WAIT_CANCEL,		(void*) gsc_wait_cancel_ioctl	},
	{ DSI16WRC_IOCTL_WAIT_STATUS,		(void*) gsc_wait_status_ioctl	},
	{ DSI16WRC_IOCTL_XCVR_TYPE,			(void*) _xcvr_type				},
	{ DSI16WRC_IOCTL_EXT_SYNC_IN,		(void*) _ext_sync_in			},
	{ -1, NULL }
};



//*****************************************************************************
static int _wait_for_ready(dev_data_t* dev, const char* reason, long ms_limit)
{
	int	ret;
	os_time_t	t1;
	os_time_t	t2;
	long	delta;

	ms_limit	*= 25;

	os_time_get(&t1);
	ret	= gsc_poll_u32(dev, ms_limit, dev->vaddr.gsc_bctlr_32, D13, D13);
	os_time_get(&t2);

	if (ret)
	{
		printf(	"%s: %s: Channels Ready not ready within %ld ms.\n",
				dev->model,
				reason,
				ms_limit);
	}
	else
	{
		delta	= os_time_delta_ms(&t1, &t2);
		printf(	"%s: %s: Channels Ready success: %ld ms.\n",
				dev->model,
				reason,
				delta);
	}

	return(ret);
}



//*****************************************************************************
static int _xlat_hw_2_sw(s32* arg, const sw_hw_t* list)
{
	int	i;
	int	ret	= 0;

	for (i = 0;; i++)
	{
		if (list[i].sw < 0)
		{
			ret	= -EINVAL;
			break;
		}

		if (list[i].hw == arg[0])
		{
			arg[0]	= list[i].sw;
			break;
		}
	}

	return(ret);
}



//*****************************************************************************
static int _xlat_sw_2_hw(s32* arg, const sw_hw_t* list)
{
	int	i;
	int	ret	= 0;

	if (arg[0] == -1)
	{
	}
	else
	{
		for (i = 0;; i++)
		{
			if (list[i].sw < 0)
			{
				ret	= -EINVAL;
				break;
			}

			if (list[i].sw == arg[0])
			{
				arg[0]	= list[i].hw;
				break;
			}
		}
	}

	return(ret);
}


